/* cdReader.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#ifdef HAVE_CDAUDIO

#include "cdReader.h"
#include "childPidWatcher.h"
#include "format.h"
#include "tableInfo.h"
#include "database.h"
#include "expand_variables.h"
#include "busyMessage.h"
#include "infoMessage.h"

#include <stdio.h>
#include <errno.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>

CDReader::
CDReader(const Dataset &data, const string &device_name) :
  _data(data),
  _device_name(device_name) 
{
  BusyMessage bm("Reading " + device_name);

  _cdrom_type = data[TI_configure][CFI_cdrom_type].get_int();
  _cdrom_load_type = data[TI_configure][CFI_cdrom_load_type].get_int();

  _has_disc = false;
  _has_names = false;
  _read_pid = 0;

  _cd_desc = cd_init_device((char *)_device_name.c_str());

  if (_cd_desc < 0) {
    if (errno == EBUSY) {
      new InfoMessage("Cannot read " + _device_name +
		      " as it is currently mounted.");
    } else {
      perror(_device_name.c_str());
    }

    return;
  }

  // Make sure the tray is closed before we try to read the CD.
  cd_close(_cd_desc);

  cd_stat(_cd_desc, &_disc_info);

  if (!_disc_info.disc_present) {
    cd_stat(_cd_desc, &_disc_info);
    if (!_disc_info.disc_present) {
      new InfoMessage("No CD in drive.");
      return;
    }
  }

  _has_disc = true;
  _has_names = false;
  _cddb_id = cddb_discid(_cd_desc);
}

const string &CDReader::
get_device_name() const {
  return _device_name;
}

bool CDReader::
query_cddb() {
  _has_names = true;

  // Sometimes cddb_read_disc_data() thinks it succeeds, but it
  // doesn't, and it doesn't change the data from last time.  To avoid
  // getting burned by this, we'll zero out the data ahead of time.

  // Actually, this doesn't seem to work.  Huh.

  memset(&_disc_data, 0, sizeof(_disc_data));
  if (cddb_read_disc_data(_cd_desc, &_disc_data) < 0) {
    perror(_device_name.c_str());
    _has_names = false;

  } else {
    // cddb_read_disc_data() might fail to find any information,
    // returning "Unknown" in all fields, but still return success.  If
    // this happens, we'll mark the names as unknown.

    if (get_title() == "Unknown" && get_artist() == "Unknown") {
      bool unknown = true;
      int num_tracks = get_num_tracks();
      for (int t = 0; t < num_tracks && unknown; t++) {
	unknown = (get_track_name(t) == "Unknown");
      }
      
      if (unknown) {
	_has_names = false;
      }
    }
  }
  return _has_names;
}


CDReader::
~CDReader() {
  if (_cd_desc >= 0) {
    // Silly me.  cd_close means close the tray, not the file descriptor.
    //cd_close(_cd_desc);
    close(_cd_desc);
    _cd_desc = -1;
  }
}

bool CDReader::
has_disc() const {
  return _has_disc;
}

string CDReader::
get_cddb_id() const {
  assert(has_disc());
  char temp[128];
  sprintf(temp, "%08x", _cddb_id);
  return temp;
}

int CDReader::
get_num_tracks() const {
  assert(has_disc());
#ifdef CDAUDIO_OLD
  return _disc_info.disc_totaltracks;
#else
  return _disc_info.disc_total_tracks;
#endif
}

bool CDReader::
is_audio_track(int track_index) const {
  assert(has_disc());
  assert(track_index >= 0 && track_index < get_num_tracks());
#ifdef CDAUDIO_OLD
  return (_disc_info.track[track_index].track_type == CDAUDIO_TRACK_AUDIO);
else
  return (_disc_info.disc_track[track_index].track_type == CDAUDIO_TRACK_AUDIO);
#endif
}

double CDReader::
get_track_length(int track_index) const {
  assert(has_disc());
  assert(track_index >= 0 && track_index < get_num_tracks());

#ifdef CDAUDIO_OLD 
  return _disc_info.track[track_index].track_length.minutes * 60 +
    _disc_info.track[track_index].track_length.seconds;
#else
  return _disc_info.disc_track[track_index].track_length.minutes * 60 +
    _disc_info.disc_track[track_index].track_length.seconds;
#endif
}

bool CDReader::
has_names() const {
  return _has_names;
}

string CDReader::
get_title() const {
  if (!has_names()) {
    return "";
  }

  return _disc_data.data_title;
}

string CDReader::
get_artist() const {
  if (!has_names()) {
    return "";
  }

  return _disc_data.data_artist;
}

string CDReader::
get_track_name(int track_index) const {
  assert(track_index >= 0 && track_index < get_num_tracks());

  if (!has_names()) {
    return "";
  }

  return _disc_data.data_track[track_index].track_name;
}

string CDReader::
get_track_artist(int track_index) const {
  assert(track_index >= 0 && track_index < get_num_tracks());

  if (!has_names()) {
    return "";
  }

#ifdef CDAUDIO_OLD 
  return "";
#else
  return _disc_data.data_track[track_index].track_artist;
#endif
}

bool CDReader::
start_reading(int track_num, const string &filename) {
  stop_reading();

  // Get the cd command, if we don't already have it.
  if (_cd_rip_command_unexpanded.empty()) {
    QueryResult qr = music_db->sql_command
      ("SELECT command FROM soundfile_convert"
       " WHERE from_type=" + DataCell(_cdrom_type) +
       " AND to_type=" + DataCell(_cdrom_load_type));
    if (qr.get_num_rows() != 0) {
      _cd_rip_command_unexpanded = qr.get_cell(0, 0).get_string();
    }
  }

  if (_cd_rip_command_unexpanded.empty()) {
    new InfoMessage("No CD Ripping command given in database.");
    return false;
  }

  VariableDef vars;
  vars["cdrom_device"] = _device_name;
  vars["track"] = format_int(track_num);
  vars["to"] = filename;

  string cd_rip_command =
    expand_variables(_cd_rip_command_unexpanded, vars);

  // We must block the SIGCHLD signal until we've connected to the
  // watch_pid event.  Otherwise, we could miss the signal (if the
  // child terminates immediately), or worse, enter a race condition
  // with the ChildPidWatcher calls.

  sigset_t sigchld_set;
  sigemptyset(&sigchld_set);
  sigaddset(&sigchld_set, SIGCHLD);
  sigprocmask(SIG_BLOCK, &sigchld_set, NULL);

  _read_pid = fork();
  if (_read_pid < 0) {
    perror("fork");
    done_reading(false);
    _read_pid = 0;
    return false;
  }

  if (_read_pid == 0) {
    // Child.
    sigprocmask(SIG_UNBLOCK, &sigchld_set, NULL);
    setpgid(0, 0);
    execl("/bin/sh", "/bin/sh", "-c", cd_rip_command.c_str(), NULL);
    cerr << "Unable to exec " << cd_rip_command << ".\n";
    _exit(1);
  }

  // Parent.
  connect_to_method(ChildPidWatcher::watch_pid(_read_pid),
		    this, &reading_pid_done);

  sigprocmask(SIG_UNBLOCK, &sigchld_set, NULL);
}

void CDReader::
stop_reading() {
  if (_read_pid > 0) {
    kill(_read_pid, SIGTERM);
  }
}

bool CDReader::
is_reading() {
  return (_read_pid != 0);
}

double CDReader::
get_reading_status() {
  return 0.0;
}

void CDReader::
eject() {
  cd_eject(_cd_desc);
}

void CDReader::
write(ostream &out) const {
  if (!has_disc()) {
    out << "No data.\n";
    return;
  }

  if (has_names()) {
    out << "CD title is " << get_title() << "\n"
	<< "Artist is " << get_artist() << "\n";
  }
  
  int num_tracks = get_num_tracks();
  for (int t = 0; t < num_tracks; t++) {
    if (is_audio_track(t)) {
      out << "Track " << t+1 << " is "
	  << get_track_length(t) << " seconds";
      if (has_names()) {
	out << " name " << get_track_name(t);
      }
      out << "\n";
    }
  }
}

void CDReader::
reading_pid_done(pid_t pid, bool normal_exit, int exit_status) {
  if (pid != _read_pid) {
    cerr << "Unexpected child " << pid << " terminated; expected " 
	 << _read_pid << "\n";
    return;
  }

  _read_pid = 0;

  done_reading(normal_exit && exit_status == 0);
}

#endif // HAVE_CDAUDIO
