/* mySQLDatabase.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#ifdef USE_MYSQL

#include "mySQLDatabase.h"
#include "mySQLQueryResult.h"

#include <assert.h>

string MySQLDatabase::_socket;

MySQLDatabase::
MySQLDatabase(const string &name, const string &host, int port,
	      const string &user, const string &passwd) {
  _connection = mysql_init(NULL);
  assert(_connection != NULL);

  const char *host_cstr = (host.empty()) ? NULL : host.c_str();
  const char *user_cstr = (user.empty()) ? NULL : user.c_str();
  const char *passwd_cstr = (passwd.empty()) ? NULL : passwd.c_str();
  const char *socket_cstr = (_socket.empty()) ? NULL : _socket.c_str();

  MYSQL *result =
    mysql_real_connect(_connection, host_cstr, user_cstr, passwd_cstr, 
		       name.c_str(), port, socket_cstr, 0);
  if (result == NULL) {
    cerr << mysql_error(_connection) << "\n\n"
	 << "Make sure MySQL is set up properly and you have created a database\n"
	 << "named '" << name << "'.\n";
    exit(1);
  }

  check_db();
}

MySQLDatabase::
~MySQLDatabase() {
  mysql_close(_connection);
  _connection = NULL;
}

QueryResult MySQLDatabase::
sql_command(string command, bool return_failures) {
  MySQLQueryResult *mqr = new MySQLQueryResult(*this);

  if (command.empty()) {
    if (!return_failures) {
      cerr << "Attempt to execute an empty query!\n";
      abort();
    }
    mqr->set_error_message("Attempt to execute an empty query.");
    return QueryResult(mqr);
  }

  int length = command.length();
  if (command[length - 1] == ';') {
    // MySQL doesn't want the trailing semicolon.
    length--;
  }
  //  cerr << command << "\n";
  int retval = mysql_real_query(_connection, command.data(), length);

  if (retval != 0) {
    mqr->set_error_message(mysql_error(_connection));
  } else {
    MYSQL_RES *result = mysql_store_result(_connection);
    if (result == NULL) {
      // mysql_store_result returns NULL if there is an error or if
      // there was no result from the query.  Determine which it is.
      char *error_msg = mysql_error(_connection);
      if (error_msg[0] != '\0') {
	mqr->set_error_message(error_msg);
      }
    } else {
      mqr->set_my_result(result);
    }
  }

  if (!mqr->success() && !return_failures) {
    cerr << "\nMySQL error\n\n"
	 << mqr->get_error_message() << "\n"
	 << command << "\n";
    abort();
  }

  return QueryResult(mqr);
}

bool MySQLDatabase::
begin() {
  return false;
}

bool MySQLDatabase::
commit() {
  return false;
}

bool MySQLDatabase::
rollback() {
  return false;
}

bool MySQLDatabase::
has_sequence(const string &name) {
  return has_table(name);
}

bool MySQLDatabase::
create_sequence(const string &name) {
  if (sql_command("CREATE TABLE " + name + " (seq INT)").success()) {
    sql_command("INSERT INTO " + name + " (seq) VALUES (0)");
  }
  return true;
}

int MySQLDatabase::
get_next_sequence(const string &name) {
  sql_command("LOCK TABLES " + name + " WRITE");
  QueryResult qr = sql_command
    ("SELECT seq+1 FROM " + name);
  assert(qr.success());
  assert(qr.get_num_rows() == 1);
  assert(qr.get_num_columns() == 1);
  int result = qr.get_cell(0, 0).get_int();
  sql_command("UPDATE " + name + " SET seq=" + DataCell(result));
  sql_command("UNLOCK TABLES");
  return result;
}

string MySQLDatabase::
concat(const string &a, const string &b) {
  return "CONCAT(" + a + ", " + b + ")";
}

bool MySQLDatabase::
has_table(const string &name) { 
  QueryResult qr = 
    sql_command("SHOW TABLES LIKE " + DataCell(name));
  if (!qr.success()) {
    cerr << "Unable to query table " << name << "\n";
    exit(1);
  }
  return (qr.get_num_rows() > 0);
}

void MySQLDatabase::
check_db() {
  // This is supposed to verify that the connection to the server is
  // still alive.  We should probably actually do this.
  return;
}

void MySQLDatabase::
extra_args(const string &args) {
}

#endif // USE_MYSQL
